//////////////////////////////////////////////////////////////////////////
//                                                                      //
// exp.cpp -                                                            //
//                                                                      //
//                              //
//        .                          //
//                                                                      //
// Copyright (c) 2001 RC Module Inc.                                    //
//                                                                      //
//  :  ..                                                 //
//                                                                      //
//   $Revision:: 1     $      $Date:: 2/14/06 8:07p    $                //
//                                                                      //
//////////////////////////////////////////////////////////////////////////

#include "math.h"
#include "auxfd.h"

//      exp

const LONG _DTLN2 = 3196577161300663914;    //  Ln(2) * 0x4000000000000000

#ifdef __cplusplus
extern "C" {
#endif

//   
//  Ln(1+1/2)*c, Ln(1+1/4)*c, Ln(1+1/8)*c, ...
//  c = 04000000000000000
LONG _DTLN[61] ={1869877770022570995l,    1029067995697975510l,    543178378744478074l,
        279581720803300556l,    141909228037287941l,    71500440275001118l,
        35888788270063140l,    17979305495762817l,    8998414598213862l,
        4501402034722952l,    2251250236762820l,    1125762490254678l,
        562915596478891l,    281466387125573l,    140735340915370l,    70368207312213l,
        35184237871787l,        17592152490069l,        8796084633611l,        0x3ffffe00001l,
        0x1fffff80000l,      0xfffffe0000l,        0x7fffff8000l,        0x3fffffe000l,    
        0x1ffffff800l,        0xffffffe00l,        0x7ffffff80l,        0x3ffffffe0l,    
        0x1fffffff8l,        0xfffffffe,     0x7fffffff,        0x3fffffff,      0x1fffffff,    
        0xfffffff,            0x7ffffff,        0x3ffffff,        0x1ffffff,      0xffffff,    
        0x7fffff,    0x3fffff,   0x1fffff,    0xfffff,        0x7ffff,      0x3ffff,    
        0x1ffff,    0xffff,        0x7fff,        0x3fff,            0x1fff,          0xfff,    
        0x7ff,        0x3ff,        0x1ff,        0xff,
        0x7f,    0x3f,    0x1f,    0xf,    0x7,    0x3,    0x1};

// exp --
//
//    
//
//      --.
//      CORDIC,    .

double exp(double f)
{

    LONG Mnt;               //   
    int Exp, Sgn;           //     
    int NewExp;             //   
    ULONG Arg;              //      
    //   double
    DstrDouble( f, Mnt, Exp, Sgn);

    //   
    if ( Exp == 0x400 && Mnt == 0 && Sgn == 0 )    //   1 +inf
        return f;

    if ( Exp == 0x400 && Mnt == 0 && Sgn == 1 )    //   2 -inf
        return 0;

    if ( Exp == -0x3ff && Mnt != 0 ){              //   3 
        return 1;
    }

    if ( Exp == 0x400 && Mnt != 0 )                //   4 NAN
        return f;

    if ( Exp > 9 && Sgn == 0 ){            //    
        if ( !_FPIEEE_Flags )
            _FPIEEE_Signal( _FPIEEE_Overflow );
        return ConstructDouble( 0, 0x400, 0, 0);
    }

    if ( Exp > 9 && Sgn == 1 ){            //    
        if ( !_FPIEEE_Flags )
            _FPIEEE_Signal( _FPIEEE_Underflow );
        return ConstructDouble( 0, -0x3ff, 0, 0);
    }

    if ( Exp < -54 )            //     
        return ConstructDouble( 0, 0, 0, 0);    

    //       ,
    //    -   
    int i = -11;          

    //      () ,  floor

    //      ,  - Arg * 2^i
    if ( Exp >= 0 ){
        Arg = Mnt << (Exp);
    }
    else if ( Exp >= -2 ){
        Arg = Mnt << 4;
        Arg = Arg >> (-Exp);
        i = -7;
    }
    else {
        Arg = Mnt << 11;
        Arg = Arg >> (-Exp);
        i = 0;
    }

    //   ,   
    LONG tmp; //        
    ULONG rez = 0x4000000000000000; //    1, 
                                    //      
    NewExp = 0;
    tmp = Arg;

    //     .     
    //  2^i * Ln(2),      2^i. ( -i)
    while ( i < 0 ){      //   
        i++;
        tmp -= _DTLN2;
        if ( tmp > 0 ){
            Arg = tmp;
            NewExp += 1 << (-i);
        }
        else            //     - 
            tmp = Arg;
        tmp <<= 1;
        Arg = tmp;
    }

    tmp >>= 1;   //    ,    
    Arg = tmp;
    //  ,      
    if ( NewExp > 0x3ff ){
        if ( Sgn == 1 ){
            //       , ret=0
            if( NewExp > 0x44f ){   
                _FPIEEE_Signal( _FPIEEE_Underflow );
                return ConstructDouble( 0, -0x3ff, 0, 0);
            }
        }
        else{
            //   , ret=+inf
            _FPIEEE_Signal( _FPIEEE_Overflow );
            return ConstructDouble( 0, 0x400, 0, 0);
        }
    }

    //   ( exp(x) = a * 2^b )    exp(-x) = ( Ln(2)-a ) * 2^(-b-1)
    if ( Sgn == 1 ){
        Arg = _DTLN2 - Arg;
        tmp = Arg;
        NewExp = -(++NewExp);
    }
    
    //      CORDIC    
    while ( i != 62 ){  //   
        tmp -= _DTLN[i++]; 
        
        //      ,
        //     ,  _DTLN[]      
        //   ,             
        if ( tmp > 0 ){
            rez += rez >> i; //   -   (1 + (0.5)^i)
            Arg = tmp;
        }
        else //     - 
            tmp = Arg;
    }
    
    if ( NewExp < -0x3fe ){    //   - 
        rez >>= (-0x3ff - NewExp) + 1;
        NewExp = -0x3ff;
    }
    
    rez >>= 9;    //     , ...
    rez = (rez >> 1) + (rez & 1); //    ,  .

    if ( rez == 0x0020000000000000 ){    
        rez = 0;
        NewExp++;
    }
    
    //   double
    return ConstructDouble( rez, NewExp, 0, 0);
}

#ifdef __cplusplus
}
#endif
