//////////////////////////////////////////////////////////////////////////
//                                                                      //
// fmod.cpp -                                                           //
//                                                                      //
//                              //
//        .                          //
//      (     Y)
//                                                                      //
// Copyright (c) 2001 RC Module Inc.                                    //
//                                                                      //
//  :  ..                                                 //
//                                                                      //
//   $Revision:: 1     $      $Date:: 2/14/06 8:07p    $                //
//                                                                      //
//////////////////////////////////////////////////////////////////////////

#include "math.h"
#include "auxfd.h"

#ifdef __cplusplus
extern "C" {
#endif

// fmod --
//
//         .  
//
double fmod(double X, double Y)
{
    /*
        fmod, fmodf, fmodl
        double fmod(double x, double y);
        float fmod(float x, float y); [C++ only]
        long double fmod(long double x, long double y); [C++ only]
        float fmodf(float x, float y); [optional]
        long double fmodl(long double x, long double y); [optional] 

        The function returns the remainder of x/y, which is defined as follows:

        If y is zero, the function either reports a domain error or simply returns zero. 
        Otherwise, if 0 <= x, the value is x - i*y for some integer i such that:
        0 <= i*|y| <= x < (i + 1)*|y| 
        Otherwise, x < 0 and the value is x - i*y for some integer i such that:
        i*|y| <= x < (i + 1)*|y| <= 0 
    */
	LONG Xm, Ym;                    //   
    LONG tmp;                       //     Xm
	int Xe, Xs;                     //      
	int Ye, Ys;                     //      
    int DeltaE;                     //   
	DstrDouble(X, Xm, Xe, Xs);      //    
	DstrDouble(Y, Ym, Ye, Ys);      //    

    if ( Xe == 0x400 ){ // X = inf, NAN
        return ConstructDouble(1, 0x400, 0, 0);
    }

    if ( Ye == 0x400 && Ym ){    // NAN
        return Ym;
    }

    if ( Ye == -0x3ff && (!Ym) ){  //  x/0
    	return ConstructDouble(1, 0x400, 0, 0);   // NAN
	}

    if (Xe > Ye+54 || Xe < -969 ){  //  969 = 0x3ff - 54 
    	return ConstructDouble(0, -0x3ff, Xs, 0);   // 0
	}

    if ( Ye > Xe ){ //       -  
        return X;
    }

    DeltaE = Xe - Ye;
    //      
    Xm <<= 10;
    Ym <<= 10;
    tmp = Xm;
    do{
        if ( (Xm-=Ym)>0 ){
            Xm <<=1;
            tmp = Xm;
            Xe--;
        }
        else{
            Xm = tmp;
            Ym >>=1;
        }
    }while (DeltaE--);

    while (!(Xm & 0x4000000000000000)){
        Xm <<=1;
        Xe--;
    }
    Xm >>= 10;

    //   double
    return ConstructDouble(Xm, Xe, Xs, 0);
}

#ifdef __cplusplus
}
#endif
